# Command line help
if { $argc != 4 } {
	  puts "Usage: ns <tcl-script> <#pod> <speed in M> <high_len> <low_util>"
      puts "E.g.: ns pareto.tcl 4 200M 100 0.1"
	  exit 1
}

# read the command line arguments
set k [lindex $argv 0]
# pareto traffic rate
set speed [lindex $argv 1]
set high [lindex $argv 2]
set low [lindex $argv 3]

set ns_ [new Simulator]

# removes all except common
remove-all-packet-headers
add-packet-header IP UDP TCP

# set up tracing
#set tracefd  [open ftree.tr w]
#$ns_ trace-all $tracefd
#set namtrace [open ftree.nam w]   
#$ns_ namtrace-all $namtrace

set bw_ 1000Mb
set pktsize 1250
set propagation_ 0.001ms
set qLimit 1000
Node set multiPath_ 1

# loop variables: p: pod, s: switch, i: host
# generating hosts
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		for {set i 0} {$i < $k/2} {incr i} {
            set h($p,$s,$i) [$ns_ node]
		}
	}
}
# generating edge switches (es)  level 1
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
           set es($p,$s) [$ns_ node]
		   $es($p,$s) qthreshold $high $low
	}
}
# generating aggregation switches (as)  level 2
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
           set as($p,$s) [$ns_ node]
		   $as($p,$s) qthreshold $high $low
	}
}
# generating core switches (cs)  level 3
for {set p 0} {$p < $k/2} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
           set cs($p,$s) [$ns_ node]
	}
}

# generating links between hosts and es
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		for {set i 0} {$i < $k/2} {incr i} {
           $ns_ duplex-link $h($p,$s,$i) $es($p,$s) $bw_ $propagation_ DropTail
		   $ns_ queue-limit $h($p,$s,$i) $es($p,$s) $qLimit
		   $ns_ queue-limit $es($p,$s) $h($p,$s,$i) $qLimit
		}
	}
}
# generating links between es and as
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		for {set i 0} {$i < $k/2} {incr i} {
           $ns_ duplex-link $es($p,$i) $as($p,$s) $bw_ $propagation_ DropTail
		   $ns_ queue-limit $es($p,$i) $as($p,$s) $qLimit
		   $ns_ queue-limit $as($p,$s) $es($p,$i) $qLimit
		}
	}
}
# generating links between as and cs
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		for {set c 0} {$c < $k/2} {incr c} {
           $ns_ duplex-link $as($p,$s) $cs($c,$s) $bw_ $propagation_ DropTail
		   $ns_ queue-limit $as($p,$s) $cs($c,$s) $qLimit
		   $ns_ queue-limit $cs($c,$s) $as($p,$s) $qLimit
		}
	}
}

# write the switch parameter to c++ space, and other initializations
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		   $es($p,$s) port-state $k 1 $p $s
	}
}
# for level 2 switches
for {set p 0} {$p < $k} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		   $as($p,$s) port-state $k 2 $p $s
	}
}
# for core switches (cs)  level 3
for {set p 0} {$p < $k/2} {incr p} {
	for {set s 0} {$s < $k/2} {incr s} {
		   $cs($p,$s) port-state $k 3 $p $s
	}
}

$ns_ rtproto DV
Agent/rtProto/DV set advertInterval  1000000

# the start time to 10 sec, this value is also set in the c++ codes
set start_time 10

# generating traffic
for {set p 0} {$p < $k} {incr p} {
    for {set s 0} {$s < $k/2} {incr s} {
        for {set i 0} {$i < $k/2} {incr i} {
            set src($p,$s,$i) [new Agent/UDP]
            $src($p,$s,$i) set packetSize_ 1400 
            set dst($p,$s,$i) [new Agent/Null]
            $ns_ attach-agent $h($p,$s,$i) $src($p,$s,$i)
            # attach the dst to host (p+1)%k, far hosts
            set q [ expr { ($p+1) % $k } ]
            $ns_ attach-agent $h($q,$s,$i) $dst($p,$s,$i)
            $ns_ connect $src($p,$s,$i) $dst($p,$s,$i)
            set e($p,$s,$i)  [new Application/Traffic/Pareto]
            $e($p,$s,$i) attach-agent $src($p,$s,$i)
            $e($p,$s,$i) set packetSize_ $pktsize
            $e($p,$s,$i) set burst_time_ 500ms
            $e($p,$s,$i) set idle_time_ 300ms
            $e($p,$s,$i) set rate_ $speed
            $e($p,$s,$i) set shape_ 1.5

            set stime [ expr $start_time + ($i + $s * $k) / 100.0 ] 
            $ns_ at $stime "$e($p,$s,$i) start"
            $ns_ at 120.8 "$e($p,$s,$i) stop"
        }
    }
}

# To initialize the hash function used in the hash threshold method
$ns_ init-crc

proc stop {} {
    global ns_ k
#global tracefd namtrace

    $ns_ print-data $k
#      $ns_ flush-trace
#      close $tracefd
#      close $namtrace
}

$ns_ at 125 "stop"
$ns_ at 125.6 "puts \"NS EXITING...\" ; $ns_ halt"

puts "Starting Simulation..."
$ns_ run
